# cython: language_level=3
import os
import shutil
from sys import exit
from sqlalchemy.orm import mapper
from sqlalchemy import text
from BioMetaDB.Accessories.ops cimport to_cstring_array, free_cstring_array
from BioMetaDB.Models.models import BaseData
from BioMetaDB.Models.functions import Record
from BioMetaDB.Config.config_manager import ConfigKeys
from BioMetaDB.Config.directory_manager import Directories
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.Exceptions.fix_exceptions import FixAssertString


def record_bad_location_none(**kwargs):
    """ Function will announce that no changes were made for the given record's file location

    :param kwargs:
    :return:
    """
    assert_kwargs_loaded(kwargs, "_id")
    return "No changes to location info for %s" % kwargs["_id"]

def record_bad_location_file(**kwargs):
    """ Function will handle issues in which a record in the database does not have a valid file

    :param kwargs:
    :return:
    """
    assert_kwargs_loaded(kwargs, "_id", "sess", "UserClass", "fix_data", "config", "table_name")
    try:
        shutil.copy(
            kwargs["fix_data"],
            os.path.join(kwargs["config"][ConfigKeys.DATABASES][ConfigKeys.db_dir], kwargs["table_name"])
        )
    except FileNotFoundError:
        return "Unable to set location for %s" % kwargs["_id"]
    except shutil.SameFileError:
        return "File in location"
    record = (kwargs["sess"]).query(kwargs["UserClass"]).filter(text("_id == '%s'" % kwargs["_id"])).first()
    setattr(record, "location", os.path.join(kwargs["config"][ConfigKeys.DATABASES][ConfigKeys.db_dir], kwargs["table_name"]))
    setattr(record, "_id", os.path.basename(kwargs['fix_data']))
    return "Info <location> set to %s for %s" % (kwargs["fix_data"], kwargs["_id"])

def file_bad_record_delete(**kwargs):
    """ Function will announce that the improperly located file was deleted

    :param kwargs:
    :return:
    """
    assert_kwargs_loaded(kwargs, "_id")
    if os.path.exists(kwargs["_id"]):
        os.remove(kwargs["_id"])
    return "File %s was deleted" % kwargs["_id"]

def file_bad_record_record(**kwargs):
    """ Function will use passed info in fix_data to assign _id to a given record in the table directory

    :param kwargs:
    :return:
    """
    assert_kwargs_loaded(kwargs, "_id", "sess", "UserClass", "fix_data")
    record = (kwargs["sess"]).query(kwargs["UserClass"]).filter(text("_id == '%s'" % kwargs["fix_data"])).first()
    setattr(record, "location", kwargs["_id"])
    return "Info <_id> set to %s for %s" % (kwargs["fix_data"], kwargs["_id"])

def record_bad_type_none(**kwargs):
    """ Function will make no changes to data_type of passed _id

    :param kwargs:
    :return:
    """
    assert_kwargs_loaded(kwargs, "_id")
    return "No changes to data_type info for %s" % kwargs["_id"]

def record_bad_type_set(**kwargs):
    """ Function will set the data_type variable for the given record to match the value passed
    in the .fix file

    :param kwargs:
    :return:
    """
    assert_kwargs_loaded(kwargs, "_id", "sess", "UserClass", "fix_data")
    record = (kwargs["sess"]).query(kwargs["UserClass"]).filter(text("_id == '%s'" % kwargs["_id"])).first()
    setattr(record, "data_type", kwargs["fix_data"])
    return "Info <data_type> set to %s for %s" % (kwargs["fix_data"], kwargs["_id"])

# def table_bad_table_delete(**kwargs):
#     return 1
#
# def table_bad_table_tsv(**kwargs):
#     return 1
#
#
# def table_bad_table_mgt(**kwargs):
#     return 1


def project_invalid_path_none(**kwargs):
    """ Function displays message that no changes were done to the invalid project path

    :param kwargs:
    :return:
    """
    assert_kwargs_loaded(kwargs, "_id")
    return "No changes to project location info for %s" % kwargs["_id"]


def project_invalid_path_path(**kwargs):
    """ Function will reset all directories stored in the config folder based on the
    working directory passed in the .fix file

    :param kwargs:
    :return:
    """
    assert_kwargs_loaded(kwargs, "_id", "config", "fix_data")
    assert os.path.exists(kwargs["fix_data"]), FixAssertString.BAD_PROJECT_DIR
    config = kwargs["config"]
    # Set new working_dir and remaining config directories
    config[ConfigKeys.DATABASES][ConfigKeys.working_dir] = kwargs["fix_data"]
    config[ConfigKeys.DATABASES][ConfigKeys.db_dir] = os.path.join(kwargs["fix_data"], "db")
    config[ConfigKeys.DATABASES][ConfigKeys.config_dir] = os.path.join(kwargs["fix_data"], "config")
    config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir] = os.path.basename(kwargs["fix_data"])
    config[ConfigKeys.DATABASES][ConfigKeys.migrations_dir] = os.path.join(kwargs["fix_data"], "migrations")
    config[ConfigKeys.DATABASES][ConfigKeys.rel_db_dir] = os.path.join(os.path.basename(kwargs["fix_data"]), "db")
    config[ConfigKeys.DATABASES][ConfigKeys.db_name] = os.path.basename(kwargs["fix_data"])
    with open(os.path.join(os.path.join(config[ConfigKeys.DATABASES][ConfigKeys.working_dir], Directories.CONFIG),
                           config[ConfigKeys.DATABASES][ConfigKeys.db_name] + ".ini"), "w") as W:
        config.write(W)
    return "Info <project location> set to %s for %s" % (kwargs["fix_data"], kwargs["_id"])


def assert_kwargs_loaded(dict kwargs_dict, *args):
    """ Check that needed kwargs were passed from calling function
    
    :param kwargs_dict: 
    :param args: 
    :return: 
    """
    cdef set needed_kwargs = set(args)
    cdef set passed_args = set(kwargs_dict.keys())
    cdef str arg
    for arg in needed_kwargs:
        assert arg in passed_args, FixAssertString.INTERNAL_BAD_KWARGS


cdef class IntegrityManager:
    def __init__(self, object config, str fix_file_name, list tables=["ALL",]):
        """ IntegrityManager will check project, table, and record-level inconsistencies
        within the project structure

        :param config:
        :param fix_file_name:
        :param tables:
        """
        self.config = config
        self.function_hash = self._create_function_hash()
        self.tables = to_cstring_array(tables)
        self.fix_file = FixFile(fix_file_name)
        self.issues_found = 0

    def __del__(self):
        """ Free all allocated memory

        :return:
        """
        free_cstring_array(self.tables)

    def initial_project_check(self):
        """ Initial check to ensure that the project path is properly set in the config file
        Exits program if failed

        :return:
        """
        cdef str key
        cdef str value
        if not os.path.exists(self.config[ConfigKeys.DATABASES][ConfigKeys.working_dir]):
            self.fix_file.write_issue(
                self.config[ConfigKeys.DATABASES][ConfigKeys.db_name],
                "PROJECT",
                "INVALID_PATH",
                "NONE"
            )
            print("Error in locating path directory, canceling remaining checks - see .fix file\n")
            exit(1)
    #
    # def table_check(self):
    #     """ Method checks if config tables have directory and presence in sql tables
    #
    #     :return:
    #     """
    #     cdef set standard_config_keys = {
    #         ConfigKeys.DATABASES.lower(),
    #         ConfigKeys.TABLES_TO_DB.lower(),
    #         ConfigKeys.TABLES_TO_ALIAS.lower(),
    #         "default",
    #     }
    #     cdef str key
    #     cdef object value
    #     cdef object engine
    #     cdef object tables_in_db = MetaData(
    #         bind=BaseData.get_engine(
    #             self.config[ConfigKeys.DATABASES][ConfigKeys.db_dir],
    #             self.config[ConfigKeys.DATABASES][ConfigKeys.db_name] + ".db"
    #         ),
    #         reflect=True).tables
    #     for key, value in self.config.items():
    #         if key.lower() not in standard_config_keys and \
    #                 stringarray_contains(self.tables, to_cstring(key.lower())) == 0:
    #             # Check that config classes exist as directories
    #             if not os.path.exists(os.path.join(
    #                     self.config[ConfigKeys.DATABASES][ConfigKeys.working_dir],
    #                     "classes",
    #                     "%s.json" % key)):
    #                 self.fix_file.write_issue(key, "TABLE", "BAD_TABLE", "DELETE")
    #                 self.issues_found = 1
    #             # Check that config classes exist in database
    #             if not key.lower() in tables_in_db:
    #                 self.fix_file.write_issue(key, "TABLE", "BAD_TABLE", "DELETE")
    #                 self.issues_found = 1

    def initialize_fix_file(self):
        self.fix_file.initialize_file()

    def record_check(self, object sess, object UserClass, str table_name):
        """ Method will check consistency of each individual record

        :param table_name:
        :param sess:
        :param UserClass:
        :return:
        """
        cdef object record
        cdef list all_records = sess.query(UserClass).all()
        cdef set records_in_table_dir = set([os.path.basename(val) for val in os.listdir(
            os.path.join(self.config[ConfigKeys.DATABASES][ConfigKeys.db_dir],
                         table_name))])
        cdef str record_full_path
        cdef str _file
        cdef set record_paths = set()
        for record in all_records:
            record_full_path = record.full_path()
            if record_full_path is not None:
                record_paths.add(os.path.basename(record_full_path))
            # Data type of record is not known
            if record.data_type == "unknown":
                self.fix_file.write_issue_with_location(
                    record._id, "RECORD", "BAD_TYPE", "NONE", table_name)
                self.issues_found = 1
            # File for record, as stored in db, does not exist
            if record_full_path is None or not os.path.exists(record_full_path):
                self.fix_file.write_issue_with_location(
                    record._id, "RECORD", "BAD_LOCATION", "NONE", table_name)
                self.issues_found = 1
        for _file in records_in_table_dir.difference(record_paths):
            # File is in db directory but not in database table
            self.fix_file.write_issue_with_location(
                _file, "FILE", "BAD_RECORD", "DELETE", table_name)
            self.issues_found = 1

    def parse_and_fix(self, bint silent):
        """ Reads .fix file and calls function for all valid issues

        :return:
        """
        self.fix_file.load_file()
        self.fix_file.read_issue()
        cdef list issues = self.fix_file.issues
        cdef int i
        cdef object return_val = ""
        cdef object sess = None, UserClass = None, engine = None, TableClass = None
        cdef str current_table_name
        engine = BaseData.get_engine(self.config[ConfigKeys.DATABASES][ConfigKeys.db_dir],
                                     self.config[ConfigKeys.DATABASES][ConfigKeys.db_name] + ".db")
        sess = BaseData.get_session_from_engine(engine)
        # Generate table class and name based on presence of alias
        # Issues are in format (_id, location, parsed_issue, fix_data)
        current_table_name = issues[0][1]
        if current_table_name != "NONE":
            TableClass = ClassManager.get_class_orm(current_table_name, engine)
            UserClass = type(current_table_name, (Record,), {})
            mapper(UserClass, TableClass)
        cdef set possible_functions = set(self.function_hash.keys())
        for i in range(self.fix_file.num_issues):
            # Only re-generate table information if needed
            current_table_name = issues[i][1]
            if current_table_name != "NONE" and issues[i][1] != current_table_name:
                current_table_name = issues[i][1]
                TableClass = ClassManager.get_class_orm(current_table_name, engine)
                UserClass = type(current_table_name, (Record,), {})
                mapper(UserClass, TableClass)
            # Clear info as needed
            elif issues[i][1] == "NONE":
                current_table_name = "NONE"
                TableClass = None
                UserClass = None
            # Run fixing function if valid
            if issues[i][2] in possible_functions:
                return_val = self.function_hash[issues[i][2]](_id=issues[i][0],
                                                              location=issues[i][1],
                                                              sess=sess,
                                                              UserClass=UserClass,
                                                              TableClass=TableClass,
                                                              config=self.config,
                                                              fix_data=issues[i][3],
                                                              engine=engine,
                                                              cfg=self.config,
                                                              silent=silent,
                                                              table_name=current_table_name)
            else:
                print("Invalid fix data passed for %s" % issues[i][0])
            # Print results of each fix
            if return_val and not silent:
                print(return_val, "\n")
        # Commit all changes to the database
        sess.commit()

    cdef dict _create_function_hash(self):
        cdef dict function_hash = {}
        function_hash["RECORD|BAD_LOCATION|NONE"] = record_bad_location_none
        function_hash["RECORD|BAD_LOCATION|FILE"] = record_bad_location_file
        function_hash["FILE|BAD_RECORD|DELETE"] = file_bad_record_delete
        function_hash["FILE|BAD_RECORD|RECORD"] = file_bad_record_record
        function_hash["RECORD|BAD_TYPE|NONE"] = record_bad_type_none
        function_hash["RECORD|BAD_TYPE|SET"] = record_bad_type_set
        # function_hash["TABLE|BAD_TABLE|DELETE"] = table_bad_table_delete
        # function_hash["TABLE|BAD_TABLE|TSV"] = table_bad_table_tsv
        # function_hash["TABLE|BAD_TABLE|MGT"] = table_bad_table_mgt
        function_hash["PROJECT|INVALID_PATH|NONE"] = project_invalid_path_none
        function_hash["PROJECT|INVALID_PATH|PATH"] = project_invalid_path_path
        return function_hash



# cython: language_level=3
import os

cdef extern from "stdio.h":
    FILE *fopen(const char *, const char *)
    int fclose(FILE *)
    size_t getline(char **, size_t *, FILE*)
    int fprintf  (FILE *stream, const char *template, ...)


cdef class FixFile:
    def __init__(self, object file_name):
        """ FixFile writes and reads .fix files from project integrity checks

        :param file_name:
        """
        self.file_name = file_name
        self.fp = None
        self.issues = []
        self.num_issues = 0

    cdef initialize_file(self):
        """ Creates fix file using template in module

        :return:
        """
        cdef object cfile = open(self.file_name, "w")
        if cfile is None:
            raise FileNotFoundError(2, "No such file or directory: %s" % self.file_name)
        cdef str py_header_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                               "fix_header.txt")
        cdef object fix_header_ptr = open(py_header_path, "rb")
        for line in fix_header_ptr:
            cfile.write(line.decode())
        self.fp = cfile
        fix_header_ptr.close()

    cdef load_file(self):
        """ Loads file generated in INTEGRITY
        
        :return: 
        """
        cdef object cfile = open(self.file_name, "r")
        if cfile is None:
            raise FileNotFoundError(2, "No such file or directory: %s" % self.file_name)
        self.fp = cfile

    cdef read_issue(self):
        cdef object line
        cdef str data_type, issue_type, fix_type, fix_data = "NONE", location = "NONE", parsed_issue
        for line in self.fp:
            while line.startswith("@"):
                line = next(self.fp)
            line = line.rstrip("\r\n")
            if line == "ISSUE:":
                line = next(self.fp)
                self.num_issues += 1
                line = line.rstrip("\r\n").split(":")
                data_type = line[0]
                _id = line[1].lstrip("\t")
                if len(line) == 3:
                    location = line[2].lstrip("\t")
                line = next(self.fp)
                line = line.rstrip("\r\n").split("\t")
                issue_type = line[0]
                fix_type = line[1]
                if len(line) == 3:
                    fix_data = line[2]
                parsed_issue = "%s|%s|%s" % (data_type, issue_type, fix_type)
                self.issues.append((_id, location, parsed_issue, fix_data,))


    def write_issue_with_location(self, str issue_id, str data_type, str issue_type, str fix_type, str location):
        """
        
        :param location: 
        :param issue_id:
        :param data_type:
        :param issue_type: 
        :param fix_type: 
        :return: 
        """
        self.fp.write("ISSUE:\n%s:\t%s:\t%s\n%s\t%s\n---\n" % (data_type, issue_id, location, issue_type, fix_type))


    def write_issue(self, str issue_id, str data_type, str issue_type, str fix_type):
        """
        
        :param issue_id:
        :param data_type:
        :param issue_type: 
        :param fix_type: 
        :return: 
        """
        self.fp.write("ISSUE:\n%s:\t%s\n%s\t%s\n---\n" % (data_type, issue_id, issue_type, fix_type))




from itertools import chain, islice
from libc.stdlib cimport malloc, free
from libc.string cimport strcmp, strlen

"""
Script holds functions for optimizing speed (e.g. chunking, splitting at line, etc)

"""
cdef extern from "Python.h":
    char* PyUnicode_AsUTF8(object unicode)


def chunk(iterable, int n):
    """ For parsing very large files

    :param n: (int) Chunk size
    :param iterable: (iter)	File iterable
    """
    iterable = iter(iterable)
    while True:
        yield chain([next(iterable)], islice(iterable, n - 1))


def read_until(str line, str delim):
    """ Return value before delimiter

    :param line: (str)	Line to read
    :param delim: (str)	Break at char (non-inclusive)
    :return str:
    """
    cdef int char_count = 0
    cdef str _c
    for _c in line:
        if _c != delim:
            char_count += 1
        else:
            break
    return line[:char_count]


def split_line_at(str line, str delim, bint return_delim=False):
    """ Return value before delimiter

    :param return_delim: (bool) Return the end of line with the delimiter
    :param line: (str)	Line to read
    :param delim: (str)	Break at char (non-inclusive)
    :return str:
    """
    cdef int char_count
    cdef str _c
    char_count = 0
    for _c in line:
        if _c != delim:
            char_count += 1
        else:
            break
    if return_delim:
        return line[:char_count], line[char_count:]
    else:
        return line[:char_count], line[char_count + 1:]


def touch(path):
    """ Create file (mimics touch in unix)

    :param path: (str)  Location to create file
    :return:
    """
    open(path, 'a').close()


def print_if_not_silent(bint silence_value, str output_text):
    """ Prints value to standard output if user-setting for "silent" is not set

    :param silence_value:
    :param output_text:
    :return:
    """
    if not silence_value:
        print(output_text)


cdef char** to_cstring_array(list list_str):
    """ Converts list of python strings to c array
    
    :param list_str: 
    :return: 
    """
    cdef char **ret = <char **>malloc(len(list_str) * sizeof(char *))
    cdef int i
    for i in range(len(list_str)):
        ret[i] = PyUnicode_AsUTF8(list_str[i])
    return ret


cdef free_cstring_array(char **cstring_array):
    """ Frees array of character arrays
    
    :param cstring_array: 
    :return: 
    """
    free(cstring_array)


cdef str to_pystring(char* cstring):
    cdef str to_return = ""
    cdef size_t cstring_length = strlen(cstring)
    cdef size_t i
    for i in range(cstring_length):
        to_return += chr(cstring[i])
    return to_return


cdef char* to_cstring(str py_string):
    """ Converts str to char*
    
    :param py_string: 
    :return: 
    """
    cdef bytes py_bytes = py_string.encode()
    cdef char* c_string = py_bytes
    return c_string


cdef free_cstring(char* cstring):
    """ Frees memory from char array
    
    :param cstring: 
    :return: 
    """
    free(cstring)


cdef int count_characters(char* cstring, const char* search_val):
    """ Counts number of occurrences of search_val
    
    :param cstring: 
    :param search_val: 
    :return: 
    """
    cdef int count = 0
    cdef size_t cstring_size = strlen(cstring)
    cdef size_t i
    for i in range(cstring_size):
        if cstring[i] == search_val[0]:
            count += 1
    return count


cdef int stringarray_contains(char** list_of_strings, char* string_to_find):
    cdef size_t list_length = sizeof(list_of_strings) / sizeof(list_of_strings[0])
    cdef size_t i
    for i in range(list_length):
        if strcmp(list_of_strings[i], string_to_find) == 0:
            return 0
    return 1



# cython: language_level=3
import os
import csv
import glob

from BioMetaDB.DBManagers.type_mapper import TypeMapper
from BioMetaDB.Models.functions import Record
from sqlalchemy.orm import mapper
from BioMetaDB.Accessories.ops import print_if_not_silent
from BioMetaDB.Serializers.count_table import CountTable


class UpdateManager:
    IGNORE = ("BaseData", "DataTypes", "FileLocations", "DBManager", "Base")
    MGMT_EXT = ".migrations.mgt"

    def __init__(self, object config_manager_object, dict class_as_dict, object session):
        self.cfg = config_manager_object
        self.class_as_dict = class_as_dict
        self.session = session

    def create_table_copy(self, str outfile_prefix, object DBClass, bint silent):
        """ Function writes a copy of data currently stored in the DB table

        :param silent:
        :param DBClass:
        :param outfile_prefix:
        :return:
        """
        print_if_not_silent(silent, " ..Creating deep copy of existing table")
        return self._write_to_file(*self._query_table(DBClass), outfile_prefix)

    @staticmethod
    def delete_old_table_and_populate(object engine, object TableClass, object UpdatedClass, object data_table, str table_name, object sess, bint silent,
                                      list ignore_fields=[]):
        print_if_not_silent(silent, " ..Deleting old table schema")
        TableClass.drop(engine)
        print_if_not_silent(silent, " ..Creating new table and filling with existing data")
        UpdatedClass.create(engine)
        records = UpdateManager.create_from_csv(data_table, {table_name: UpdatedClass}, ignore_fields)
        cdef object record
        for record in records[table_name]:
            sess.add(record)
        sess.commit()

    def _query_table(self, object DBClass):
        """ Queries single table in database based on class

        :param DBClass:
        :return: (Dict[str, List[db_objects]])      DB data
        """
        cdef list data = self.session.query(DBClass).all()
        # Return list of columns and dict with name of table as key and queried data as value
        try:
            return {self.cfg.table_name: [col for col in data[0].keys() if col != "_sa_instance_state"], }, \
                   {self.cfg.table_name: data, }
        except IndexError:
            return {self.cfg.table_name: self.class_as_dict.keys(), }, {
                self.cfg.table_name: data, }

    def _write_to_file(self, dict cols, dict data, str outfile_prefix):
        """ Protected method that writes or prints query data. Called when write_to_file or print_all is called.

        :param outfile_prefix: (str) Prefix to give stored data
        :param cols: columns output from cls.query_all(param)
        :param data: query_data output from cls.query_all(param)
        :return:
        """
        # Iterate over tables
        cdef str outfile_path = UpdateManager._check_migration_file_number(
            os.path.join(self.cfg.migrations_dir, outfile_prefix + UpdateManager.MGMT_EXT))
        cdef str table_name
        cdef object col_list
        cdef object record
        W = open(os.path.join(self.cfg.migrations_dir, outfile_path), "w")
        for table_name in data.keys():
            # List of columns in table
            col_list = cols[table_name]
            # Write name of table as first line
            W.write('"Record","' + table_name + '"' + "," + "\n")
            # Write comma-separated names of each column
            W.write(",".join(['"{}"'.format(col) for col in col_list]) + "\n")
            # Iterate over every entry
            for record in data[table_name]:
                # Write each entry by column
                # W.write(','.join(['"{}"'.format(entry[col]) for col in col_list]) + "\n")
                W.write(','.join([str(getattr(record, col)) for col in col_list]) + "\n")
                # W.write(','.join([str(ent) for ent in entry]) + "\n")
            W.write("\n")
        W.close()
        return W.name

    @staticmethod
    def _check_migration_file_number(str simple_migration_path):
        """ Protected method will determine the most recent migration and will write new copy

        :param simple_migration_path:   (str) template migration string as ########.migrations.mgt
        :return:
        """
        cdef str path_prefix
        cdef list possible_migrations
        if os.path.isfile(simple_migration_path):
            possible_migrations = sorted(glob.glob("%s*" % simple_migration_path.rstrip(UpdateManager.MGMT_EXT)))
            if len(possible_migrations) > 1:
                possible_migrations.pop(-1)
            if len(possible_migrations) == 1:
                path_prefix = os.path.basename(possible_migrations[0])
            else:
                path_prefix = os.path.basename(possible_migrations[-1])
            path_prefix = path_prefix.rstrip(UpdateManager.MGMT_EXT)
            # First rewrite:    #########.migrations.mgt
            if len(path_prefix) == 8:
                return path_prefix + ".001" + UpdateManager.MGMT_EXT
            # Second rewrite:   ########.###.migrations.mgt
            if len(path_prefix) == 12:
                path_prefix, current_version = path_prefix.split(".")
                return path_prefix + "." + "{:03d}".format(int(current_version) + 1) + UpdateManager.MGMT_EXT
        return simple_migration_path

    @staticmethod
    def _load_from_csv(str csv_file):
        """ Protected method to load csv data and dictionaries of columns and data

        :param csv_file:
        :return Tuple[Dict[str, List[str]], Dict[str, List[List[str]]]]:
        """
        cdef dict tables = {}
        cdef object _csv_file = open(csv_file, newline='')
        cdef object csv_reader = csv.reader(_csv_file, delimiter=",")
        cdef dict cols = {}
        cdef str dept
        # Load each table data into cols and tables variables
        for row in csv_reader:
            if row[0] == "Record":
                dept = row[1]
                tables[dept] = []
                row = next(csv_reader)
                cols[dept] = row
                row = next(csv_reader)
                while row != "\n" and row:
                    tables[dept].append(row)
                    row = next(csv_reader)
        return cols, tables

    @staticmethod
    def create_from_csv(object csv_file, dict tables, list ignore_fields):
        """ Create database objects using new schema.
        This is to be completed once all data has been backed up from the server

        :param ignore_fields:
        :param tables:
        :param csv_file:
        :return:
        """
        # Load data from .csv file
        cdef dict cols
        cdef dict csv_data
        cdef object db_object
        cdef int i
        cols, csv_data = UpdateManager._load_from_csv(csv_file)
        cdef dict db_objects = {}
        cdef str name
        cdef object Table
        for name, Table in tables.items():
            db_objects[name] = []
            # Create DB object using json data stored in file
            DBClass = type(name, (Record,), {})
            mapper(DBClass, Table)
            if name in csv_data.keys():
                # Skip certain fields
                # Useful for editing columns
                for entry in csv_data[name]:
                    db_object = DBClass()
                    for i in range(len(cols[name])):
                        if cols[name][i] not in ignore_fields:
                            setattr(db_object, cols[name][i], CountTable._try_return(entry[i]))
                    # Add to list of objects to commit
                    db_objects[name].append(db_object)
        return db_objects



# cython: language_level=3
import os
class TSVJoiner:
    def __init__(self, str tsv_file_path = ""):
        self.header = set()
        self.data = {}

        if os.path.exists(tsv_file_path):
            self.read_tsv(tsv_file_path)

    def _join_header(self, list header):
        cdef str _h
        for _h in header:
            self.header.add(_h)

    def read_tsv(self, str tsv_file_path):
        assert os.path.exists(tsv_file_path), "%s does not exist" % tsv_file_path
        cdef object R = open(tsv_file_path, "r")
        header = next(R).rstrip("\r\n").split("\t")
        cdef int header_len = len(header)
        data = None
        cdef str _line
        self._join_header(header[1:])
        for _line in R:
            line = _line.rstrip("\r\n").split("\t")
            data = self.data.get(line[0], None)
            if data is None:
                self.data[line[0]] = {}
            for i in range(1,header_len):
                self.data[line[0]][header[i]] = line[i]
        R.close()

    def write_tsv(self, str out_tsv_path):
        W = open(out_tsv_path, "w")
        header_list = list(self.header)
        cdef str head
        cdef str _id
        _out = ""
        W.write("ID\t")
        for head in header_list:
            _out += head + "\t"
        W.write(_out[:-1] + "\n")
        for _id in self.data.keys():
            _out = _id + "\t"
            for head in header_list:
                _out += self.data[_id].get(head, "None") + "\t"
            W.write(_out[:-1] + "\n")
        W.close()



# cython: language_level=3

import argparse
from argparse import RawTextHelpFormatter


class ArgParse:

    def __init__(self, arguments_list, description, *args, **kwargs):
        """ Class for handling parsing of arguments and error handling

        Example:

        args_list = [
            [["required_argument"],
                {"help": "Help string for argument"}],
            [["-o", "--optional"],
                {"help": "Optional argument", "default": "None"}],
            [["-r", "--required"],
                {"help": "Required argument", "required": "True"}]
        ]

        ap = ArgParse(args_list, description="Sample:\tSample program")

        ## Now you can access as ap.args.required_argument, ap.args.optional, and ap.args.required
        ## This script will handle requirement checking, and will not allow the script to launch unless required flags
            are set.
        ## Note that you CANNOT use '-' in the names of arguments!
        ## Note that any other constructor argument that argparse.ArgumentParser() takes will also be used

        Ensure that the final value in the inner list does not have '-' characters
        Include "require": True in inner dictionary to treat arg as required

        :param arguments_list: List[List[List[str], Dict[str, str]]]
        """
        self.arguments_list = arguments_list
        self.args = []
        # Instantiate ArgumentParser
        self.parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter, description=description,
                                              *args, **kwargs)
        # Add all arguments stored in self.arguments_list
        self._parse_arguments()
        # Parse arguments
        try:
            self.args = self.parser.parse_args()
        except:
            exit(1)

    def _parse_arguments(self):
        """ Protected method for adding all arguments stored in self.arguments_list
            Checks value of "require" and sets accordingly

        """
        for args in self.arguments_list:
            # Be default names argument by last value passed in inner list
            self.parser.add_argument(*args[0], **args[1])

    @staticmethod
    def description_builder(str header_line, dict help_dict, dict flag_dict):
        assert set(help_dict.keys()) == set(flag_dict.keys()), "Program names do not match in key/help dictionaries"
        to_return = header_line + "\n\nAvailable Programs:\n\n"
        programs = sorted(flag_dict.keys())
        for program in programs:
            to_return += program + ": " + help_dict[program] + "\n\t" + \
                         "\t(Req: {})".format(" --" + " --".join(flag_dict[program])) + "\n"
        to_return += "\n"
        return to_return



# cython: language_level=3
cdef class CountTable:
    cdef public dict file_contents
    cdef public list header

    def __init__(self, str file_name):
        """ Class parses count tables from KEGG workflow

        :param file_name: (str)	count table output (from KEGG workflow)
        """
        self.file_contents = self._read_contents(file_name)

    def _read_contents(self, str file_name):
        """ Protected method to read file and return set of matches from file


        :param file_name: KEGG workflow output table
        :return Dict[str, namedtuple(Genome)]:
        """
        cdef dict contents_in_file = {}
        cdef list line
        cdef bytes _line
        with open(file_name, "rb") as R:
            # Save first line, removing empty tabs that may exist on header line
            self.header = [val for val in R.readline().decode().rstrip("\r\n").split("\t")[1:] if val != ""]
            for _line in R:
                # ID/key is first item on line
                line = _line.decode().rstrip("\r\n").split("\t")
                contents_in_file[line[0]] = line[1:]
        return contents_in_file

    def get_line(self, str genome_id, bint endline=False):
        """ Returns line from count table based on genome id

        :param genome_id: (str)	ID of genome
        :param endline: (bool)  Adds newline character
        :return str:
        """
        if endline:
            return "\t".join(self.file_contents[genome_id]) + "\n"
        else:
            return "\t".join(self.file_contents[genome_id])

    def get_header_line(self, bint endline=False, bint start_field=False):
        """ Returns header as string

        :param endline: (bool)  Add newline character
        :return str:
        """
        cdef str return_string = "\t".join(self.header)
        if endline:
            return_string += "\n"
        elif start_field:
            return_string = "ID\t" + return_string
        return return_string

    def get_at(self, str genome_id, int location):
        """ Returns value from count table based on genome id and location
        Note that, if header indices are [0,1,2,3,...],
        then the locations are gathered by [1,2,3,4,...] indices
        
        :param genome_id: (str) ID of genome
        :param location: (int)      Location in line (first column is location 0)
        :return str: 
        """
        return CountTable._try_return(self.file_contents[genome_id][location])

    def get_line_with_taxa(self, str genome_id, str taxa):
        """ Returns genome id with added string (typically taxonomy) appended

        :param genome_id: (str)	ID of genome
        :param taxa:	(str)	Requested taxa (e.g. family, phylum, etc)
        :return str:
        """
        cdef list content = self.file_contents[genome_id]
        content[0] += "." + taxa
        return "\t".join(content)

    @staticmethod
    def _try_return(object value):
        """ Attempts to parse value as integer and then as float. Simply returns if fails

        :param value: (str) Value from count table
        :return:
        """
        if value == "True":
            return True
        elif value == "False":
            return False
        elif value == 'None':
            return None
        try:
            if "." not in value:
                return int(value)
            raise ValueError
        except ValueError:
            try:
                if "." in value or "e" in value or "E" in value:
                    return float(value)
                raise ValueError
            except ValueError:
                if value and value != '':
                    return value
                else:
                    return None

    @staticmethod
    def _try_return_type(object value):
        """ Attempts to parse value as integer and then as float. Simply returns if fails

        :param value: (str) Value from count table
        :return:
        """
        if value in ("True", "False"):
            return bool
        try:
            if "." not in value:
                val = int(value)
                return int
            raise ValueError
        except ValueError:
            try:
                if "." in value or "E" in value or "e" in value:
                    val = float(value)
                    return float
                raise ValueError
            except ValueError:
                if value != '':
                    return str
                else:
                    return None



import os
from BioMetaDB.Accessories.bio_ops import BioOps

"""
Script holds superclass from which all user-created classes will inherit

"""


class Record:

    def __init__(self):
        self._id = None
        self.location = None
        self.temp_filename = None

    def __del__(self):
        self.clear_temp()

    def full_path(self):
        """ Method returns full path of bio storage location

        :return str:
        """
        cdef str file_path = None
        if os.path.isfile(self.location + "/" + self._id):
            file_path = self.location + "/" + self._id
        elif os.path.isfile(self.location + "/" + self._id + ".gz"):
            file_path = self.location + "/" + self._id + ".gz"
        return file_path

    def __repr__(self):
        """

        :return str:
        """
        # Print sorted attributes, excluding instance state object info
        cdef tuple first_vals = ("_id", "data_type", "location")
        cdef tuple first_cor_vals = ("ID", "Data Type", "File Location")
        cdef str key
        cdef dict attrs = {key: val for key, val in self.__dict__.items() if key != "_sa_instance_state"
                 and key not in first_vals and key != "id"}
        cdef list sorted_keys = sorted(list(attrs.keys()))
        cdef list reorder = ["domain", "phylum", "_class", "_order", "family", "genus", "species"]
        cdef bint is_evaluation = True
        for ord in reorder:
            if ord in sorted_keys:
                continue
            is_evaluation = False
        if is_evaluation:
            for ord in reorder:
                sorted_keys.remove(ord)
            sorted_keys = reorder + sorted_keys
        cdef int i, longest_key = max([len(key) for key in attrs])
        # Pretty formatting
        summary_string = ("*" * (longest_key + 30)) + "\n"
        for i in range(len(first_vals)):
            if len(self.__dict__[first_vals[i]]) > 50:
                summary_string += "\t{:>{longest_key}}\t{:<12.47s}...\n".format(
                    first_cor_vals[i],
                    self.__dict__[first_vals[i]],
                    longest_key=longest_key)
            else:
                summary_string += "\t{:>{longest_key}}\t{:<12s}\n".format(
                    first_cor_vals[i],
                    self.__dict__[first_vals[i]],
                    longest_key=longest_key)
        summary_string += "\n\t{:>{longest_key}}\n\n".format(
            "Column Name",
            longest_key=longest_key
        )
        # Get all columns
        for key in sorted_keys:
            attr = attrs.get(key, "None")
            if type(attr) == str:
                val = attrs[key]
                if len(val) > 50:
                    val = val[:47] + "..."
                summary_string += "\t{:>{longest_key}}\t{:<50s}\n".format(key, val, longest_key=longest_key)
            elif type(attr) == bool:
                summary_string += "\t{:>{longest_key}}\t{:<50s}\n".format(key, str(attr), longest_key=longest_key)
            elif type(attr) == type(None):
                summary_string += "\t{:>{longest_key}}\t{:<50s}\n".format(key, "None", longest_key=longest_key)
            else:
                summary_string += "\t{:>{longest_key}}\t{:<50.3f}\n".format(
                    key,
                    attr,
                    longest_key=longest_key)
        summary_string += ("-" * (longest_key + 30) + "\n")
        return summary_string

    def write_temp(self):
        """ Method writes file contents to temporary file and stores file name as instance attribute

        """
        cdef str filename = self._id + ".tmp." + self.data_type
        cdef object W = open(filename, "wb")
        with open(self.full_path(), "rb") as R:
            W.write(R.read())
        W.close()
        self.temp_filename = filename

    def clear_temp(self):
        """ Method for deleting instance temporary file

        """
        if getattr(self, "temp_filename", None) is not None:
            os.remove(self.temp_filename)
            self.temp_filename = None

    def records(self):
        """ Method returns contents of record's file as a list of BioPython SeqRecord objects

        :return List[SeqRecord]
        """
        cdef str data_type = BioOps.get_type(self.full_path())
        # Parses large file if needed (chunk at 10000 lines)
        return BioOps.parse_large(self.full_path(), data_type)



# cython: language_level=3
import os
import json
import shutil
from datetime import datetime
from string import punctuation
from sqlalchemy.orm import mapper
from sqlalchemy import Column, Table, Integer, String, MetaData
from BioMetaDB.Models.models import BaseData
from BioMetaDB.Accessories.bio_ops import BioOps
from BioMetaDB.Models.functions import Record
from BioMetaDB.DBManagers.type_mapper import TypeMapper
from BioMetaDB.Accessories.ops import print_if_not_silent
from BioMetaDB.Config.directory_manager import Directories
from BioMetaDB.DBManagers.update_manager import UpdateManager

"""
Script holds ClassManager, which links together database class generation,
table creation, and class loading

"""

unusable_punctuation = set(punctuation) - set("_")
num_to_chars = {"1": "one", "2": "two", "3": "three", "4": "four", "5": "five",
                "6": "six", "7": "seven", "8": "eight", "9": "nine"}


class ClassManager:
    default_data = {
        "Float": None,
        "Integer": None,
        "String": None,
        "VARCHAR": None,
        "Boolean": None,
    }

    @staticmethod
    def create_initial_table_in_db(str db_name, str working_dir, str table_name, dict data_types, bint silent, bint initial=True):
        """ Creates initial database and tables

        :param silent: (bool)
        :param initial: (bool)      Initial table or not
        :param db_name: (str)       Name of database to create
        :param working_dir: (str)   Path to project folder
        :param table_name: (str)    Name of table to create
        :param data_types: (Dict[str, str) Attributes of class as dictionary
        :return:
        """
        cdef object engine
        cdef object metadata, t
        cdef str db_dir = os.path.join(working_dir, Directories.DATABASE)
        cdef str table_dir = os.path.join(db_dir, table_name)
        cdef str classes_dir = os.path.join(working_dir, Directories.CLASSES)
        print_if_not_silent(silent, " ..Generating table class from data file")
        data_types = ClassManager.correct_dict(data_types)
        engine = None
        if not initial:
            engine = BaseData.get_engine(db_dir, db_name + ".db")
        t, metadata = ClassManager.generate_class(table_name, data_types, db_dir, db_name, table_dir, engine=engine)
        # Add functions for accessing to table
        print_if_not_silent(silent, " ..Creating tables")
        if not initial:
            t.create(bind=engine)
        else:
            metadata.create_all()
        classes_out_file = os.path.join(classes_dir, table_name + ".json")
        print_if_not_silent(silent, " ..Saving table data as JSON to %s" % classes_out_file)
        ClassManager.write_class(data_types, classes_out_file)

    @staticmethod
    def populate_data_to_existing_table(str table_name, object count_table_object, object config, object genome_files_to_add, str directory_name,
                                        bint silent, str alias=None):
        """ Method will load data from CountTable into database

        :param silent:
        :param directory_name:
        :param alias:
        :param genome_files_to_add:
        :param count_table_object:
        :param table_name:
        :param config: (object)  ConfigManager object
        :return:
        """
        # cdef object engine, sess
        cdef set ids_to_add
        cdef set table_class_attrs_keys
        cdef set data_file_attrs_keys
        cdef dict combined_attrs
        cdef list to_add = []
        cdef int existing_records = 0
        cdef int new_records = 0
        cdef int new_records_no_files = 0
        cdef int new_records_no_data_type = 0
        cdef str _id_
        cdef object metadata, UpdatedDBClass
        print_if_not_silent(silent, "\nPopulating schema")
        print_if_not_silent(silent, " ..Loading database")
        engine = BaseData.get_engine(config.db_dir, config.db_name + ".db")
        sess = BaseData.get_session_from_engine(engine)
        print_if_not_silent(silent, " ..Collecting class data")
        TableClass = ClassManager.get_class_orm(table_name, engine)
        print_if_not_silent(silent, " ..Determining updates")
        table_class_attrs_keys = set(ClassManager.correct_iterable(ClassManager.get_class_as_dict(config).keys()))
        if count_table_object is not None:
            data_file_attrs_keys = set(ClassManager.correct_iterable(
                key for key in count_table_object.header if key not in ("", " ", "#")))
        else:
            data_file_attrs_keys = set()
        # Get combined values for writing to final JSON file
        try:
            combined_attrs = {**ClassManager.correct_dict(ClassManager.get_class_as_dict(config)),
                              **ClassManager.correct_dict(TypeMapper.get_translated_types(count_table_object,
                                                                                          TypeMapper.py_type_to_string))}
        except AttributeError:
            combined_attrs = ClassManager.correct_dict(ClassManager.get_class_as_dict(config))
        # Update manager
        # Initialize with DB class name, ConfigManager instance
        # Will create csv file of all existing data,
        # If differences found between what is in database table and what is in datafile
        if len(table_class_attrs_keys - data_file_attrs_keys) != 0 or len(data_file_attrs_keys - table_class_attrs_keys) != 0 :
            print_if_not_silent(silent, "\n!! New column data detected, calling update manager !!")
            update_manager = UpdateManager(config, ClassManager.get_class_as_dict(config), sess)
            table_copy_csv = update_manager.create_table_copy(datetime.today().strftime("%Y%m%d"), TableClass, silent)
            print_if_not_silent(silent, " ..Combining existing columns with new headers")
            UpdatedDBClass, metadata = ClassManager.generate_class(config.table_name,
                                                                   combined_attrs,
                                                                   config.db_dir,
                                                                   config.db_name,
                                                                   config.table_dir)
            ClassManager.write_class(combined_attrs, config.classes_file)
            # config.update_config_file(table_name)
            UpdateManager.delete_old_table_and_populate(engine, TableClass, UpdatedDBClass, table_copy_csv, table_name,
                                                         sess, silent)
            TableClass = UpdatedDBClass
            print_if_not_silent(silent, " ..Complete!\n")
        UserClass = type(alias or table_name, (Record,), {})
        mapper(UserClass, TableClass)
        try:
            corrected_header = ClassManager.correct_iterable(count_table_object.header)
        except AttributeError:
            corrected_header = None
        print_if_not_silent(silent, "\nGathering data by record:")
        # Or operation on sets to get all ids to add
        try:
            ids_to_add = set(count_table_object.file_contents.keys()) | \
                         set([os.path.splitext(_file)[0] for _file in genome_files_to_add if _file != "" and os.path.splitext(_file)[1] == ".gz"])
        except AttributeError:
            ids_to_add = set(count_table_object.file_contents.keys())
        if '' in ids_to_add:
            ids_to_add.remove('')
        for _id_ in ids_to_add:
            print_if_not_silent(silent, " ...Checking for record %s" % _id_)
            record = sess.query(UserClass).filter_by(_id=_id_).first()
            if record:
                print_if_not_silent(silent, " ....Record exists, updating with new values\n")
                existing_records += 1
                # Add values in the table that are already in the database class
                if corrected_header:
                    for attr in corrected_header:
                        try:
                            setattr(record, attr, count_table_object.get_at(_id_, corrected_header.index(attr)))
                        except KeyError:
                            setattr(record, attr, None)
            else:
                if directory_name != "None":
                    print_if_not_silent(silent, " ....Moving new record from %s to %s" % (directory_name,
                                                                                          config.rel_db_dir))
                    if os.path.isfile(os.path.join(directory_name, _id_)):
                        shutil.copy(os.path.join(directory_name, _id_),
                                    os.path.join(config.table_dir, _id_))
                        new_records += 1
                    elif os.path.isfile(os.path.join(directory_name, _id_ + ".gz")):
                        shutil.copy(os.path.join(directory_name, _id_ + ".gz"),
                                    os.path.join(config.table_dir, _id_ + ".gz"))
                        new_records += 1
                    else:
                        new_records_no_files += 1
                else:
                    new_records_no_files += 1
                db_object = UserClass()
                print_if_not_silent(silent, " ....Setting values for record %s in database\n" % _id_)
                setattr(db_object, "_id", _id_)
                setattr(db_object, "location", os.path.join(config.db_dir, table_name))
                try:
                    setattr(db_object, "data_type", BioOps.get_type(_id_))
                except KeyError:
                    setattr(db_object, "data_type", "unknown")
                    new_records_no_data_type += 1
                if corrected_header:
                    for attr in corrected_header:
                        try:
                            setattr(db_object, attr, count_table_object.get_at(_id_, corrected_header.index(attr)))
                        except KeyError:
                            setattr(record, attr, None)
                to_add.append(db_object)
        for val in to_add:
            sess.add(val)
        sess.commit()
        print_if_not_silent(silent, " %i existing record(s) updated" % existing_records)
        print_if_not_silent(silent, " %i new record(s) added" % new_records)
        print_if_not_silent(silent, " %i new record(s) without data files added" % new_records_no_files)
        print_if_not_silent(silent, "  %i new record(s) did not have a valid file extension\n" % new_records_no_data_type)
        print_if_not_silent(silent, "Complete!\n")
        return combined_attrs

    @staticmethod
    def write_class(dict data_types, str class_output_file):
        """ Member writes python class as json object to file for loading

        :param data_types:
        :param class_output_file:
        :return:
        """
        with open(class_output_file, "w") as W:
            json.dump(data_types, W)

    @staticmethod
    def get_class_as_dict(object cfg):
        """ Primary method for loading a class from the database

        :param cfg:
        :return:
        """
        with open(cfg.classes_file, "r") as json_data:
            class_as_dict = dict(json.load(json_data))
        return class_as_dict

    @staticmethod
    def get_class_orm(str table_name, object engine):
        """ Method for loading a class' ORM from the database

        :param engine:
        :param table_name:
        :return:
        """
        cdef object metadata = MetaData(bind=engine, reflect=True)
        return metadata.tables[table_name]

    @staticmethod
    def get_class(str table_name, object engine):
        """ Primary method for loading a class from the database

        :param alias:
        :param engine:
        :param table_name:
        :return:
        """
        cdef object metadata = MetaData(bind=engine, reflect=True)
        UserClass = type(table_name, (Record,), {})
        mapper(UserClass, metadata.tables[table_name])
        return UserClass

    @staticmethod
    def generate_class(str table_name, dict class_as_dict, str db_dir, str db_name, str table_dir, object metadata=None, object engine=None):
        """ Method generates class from dictionary and returns class

        :param metadata:
        :param table_name: (str)    Name of db table
        :param class_as_dict: (Dict[str, str]) Attributes of class as dictionary
        :param db_dir: (str)    Path to database directory
        :param db_name: (str)   Name of database
        :param table_dir: (str) Path to folder holding table/class files
        :return:
        """
        metadata = metadata or MetaData(bind=engine or BaseData.get_engine(db_dir, db_name), reflect=True)
        cdef object t = Table(table_name, metadata, Column('id', Integer, primary_key=True),
                   Column("_id", String, unique=True, index=True),
                   Column("data_type", String, index=True),
                   Column("location", String, index=True),
                   *(Column(key, TypeMapper.string_to_loaded_sql_type[value], index=True,
                           default=ClassManager.default_data[value]) for
                    key, value in class_as_dict.items() if key not in ("_id", "data_type", "location")))
        setattr(t, "basedir", table_dir)
        # for fxn in accessing_functions.keys():
        #     setattr(t, fxn, types.MethodType(accessing_functions[fxn], t))
        return t, metadata

    @staticmethod
    def correct_dict(dict class_as_dict):
        """

        :param class_as_dict:
        :return:
        """
        cdef dict corrected_dict = {}
        cdef str key
        cdef str new_key
        cdef str bad_char
        for key in class_as_dict.keys():
            new_key = str(key).lower()
            for bad_char in unusable_punctuation:
                new_key = new_key.replace(bad_char, "")
            new_key = new_key.replace(" ", "_")
            if new_key[0] in num_to_chars.keys():
                new_key = num_to_chars[new_key[0]] + new_key[1:]
            corrected_dict[new_key] = class_as_dict[key]
        return corrected_dict

    @staticmethod
    def correct_iterable(iterable):
        """ Method will make each value in iterable SQL-safe

        :param iterable:
        :return:
        """
        cdef list new_iter = []
        cdef str _iter
        cdef str new_val
        cdef str bad_char
        for _iter in iterable:
            new_val = _iter.lower()
            for bad_char in unusable_punctuation:
                new_val = new_val.replace(bad_char, "")
            new_val = new_val.replace(" ", "_")
            if new_val[0] in num_to_chars.keys():
                new_val = num_to_chars[new_val[0]] + new_val[1:]
            new_iter.append(new_val)
        return new_iter



# cython: language_level=3
import os
from Bio import SeqIO
from BioMetaDB.Accessories.ops import chunk
from BioMetaDB.Indexers.index_extensions import IndexExtensions


class BioOps:
    @staticmethod
    def get_type(str file_name):
        """ Method for parsing file extension to determine data type
        Uses dictionary of valid extensions to return "fasta" or "fastq"

        :param file_name: (str)	User-passed name of file
        """
        cdef str _f, file_type
        _f, file_type = os.path.splitext(file_name)
        if file_type == ".gz":
            _f, file_type = os.path.splitext(_f)
        return BioOps.get_corrected_data_format(file_type.strip("."))

    @staticmethod
    def calculate_phred(list scores):
        """ Determines phred quality scores based on presence of lowercase letters

        :param scores: (List[str])	List of nucleotide scores
        :return List[int]:
        """
        cdef bint is_phred_33 = True
        cdef str score
        for score in scores:
            # Determines phred based on presence of lowercase characters
            if score.islower():
                is_phred_33 = False
                break
        if is_phred_33:
            return [ord(score) - 33 for score in scores]
        else:
            return [ord(score) - 64 for score in scores]

    @staticmethod
    def get_corrected_data_format(str dt):
        """ Protected method returns corrected data type to make it easier on user

        :param dt: (str)	Inferred data type based on file extension
        :return str:
        """
        cdef dict avail_types = {
            "fasta": "fasta",
            "fastq": "fastq",
            "fna": "fasta",
            "aln": "fasta",
            "faa": "fasta",
            "fa": "fasta",
            "fq": "fastq",
            "protein": "fasta",
        }
        # Attempt simple return
        try:
            return avail_types[dt]
        except KeyError as e:
            # See if is an indexed file
            try:
                return IndexExtensions.match["." + dt]
            except KeyError:
                print("Invalid data type: {}\nConfirm correct file extension".format(e))
                raise KeyError

    @staticmethod
    def parse_large(str file_name, str data_type, int batch_size=10000):
        """

        :param file_name:	(str)	Name of fasta or fastq file
        :param data_type:	(str)	Fasta or fastq
        :param batch_size:	(int)	Size of batch for reading in, default 10000
        """
        # Create iterator
        record_iter = SeqIO.parse(open(file_name), data_type)
        # Return complete records
        return [list(batch) for batch in chunk(record_iter, batch_size)][0]



# cython: language_level=3
import traceback

"""
Script handles program calls and minimizes the amount of code that needs to be written

"""


class ProgramCaller:
    def __init__(self, programs, flags, errors, _help, classes={None: None}):
        """ ProgramCaller is used for when a script has multiple possible sub-programs available.
            Wrapper functionality is available to ensure database or file system integrity

            Example:
                from arg_parse import ArgParse
                from program_caller import ProgramCaller

                def add_file(fasta_file, data_type):
                    ...

                def remove_file(fasta_file):
                    ...

                if __name__ == "__main__":
                    programs = {	"ADD":			add_file,
                                    "REMOVE":       remove_file
                    }
                    flags = {		"ADD":			("fasta_file", "data_type"),
                                    "REMOVE":       ("fasta_file",),
                    }
                    errors = {		FileNotFoundError:		"Incorrect database called",
                    }
                    _help = {	    "ADD":		"Program for adding files to database",
                                    "REMOVE":   "Program for removing files from db"
                    }

                    args_list = [
                        [["-f", "--fasta_file"],
                            {"help": "File to pass", "default": "None"}],
                        [["-d", "--data_type"],
                            {"help": "Type of data (fasta or fastq)", "require": True}],
                    ]

                    ap = ArgParse(args_list,
                                description=ArgParse.description_builder("dbdm:\tManaging database operations.",
                                    _help, flags))
                    pc = ProgramCaller(classes=classes, programs=programs, flags=flags, errors=errors, _help=_help)

                    pc.run(ap.args)

                ## Note that the names of the arguments that each function takes are THE SAME as
                ## the arguments that are passed from ArgParse

        :param classes: (Dict[str, class])		Class mappings
        :param programs: (Dict[str, callable])	Program name: function mapping
        :param flags: (Dict[str, tuple])		Flag mapping program_name: (needed flags,)
        :param errors: (Dict[str, str])			Error string to print by error type
        """
        self.classes = classes
        self.programs = programs
        self.flags = flags
        self.errors = errors
        self.errors["INCORRECT_FLAGS"] = "Pass all necessary flags:"
        self._help = _help

    def error(self, error_type, program, message):
        """ Function called to handle error

        :param message: (str)       Message passed from exception
        :param error_type: (str)	Key for type of error
        :param program: (str)	Name of program that was called
        """
        if type(error_type) == str:
            if "flags" in error_type or "FLAGS" in error_type:
                print()
                print(" ERROR " + self.errors["INCORRECT_FLAGS"], "--" + " --".join(self.flags[program]))
        else:
            print()
            print(" ERROR " + self.errors[error_type] + "\t" + str(message))
            exit(1)
        if program in self._help.keys():
            print()
            print(" " + program + ": " + self._help[program])
            print(traceback.print_exc())
        print()
        exit(1)

    def mapper(self, program, debug=False):
        """ Function maps the name of a program with its given function and returns function call
            Catches errors that may arise and returns user-defined response

        :param debug: (bool)    Set debug options to display full error Traceback
        :param program: (str)	Name of program to use in programs dict to call function
        :return function:
        """
        if not debug:
            try:
                return self.programs[program](**self.flags_dict)
            except AssertionError as e:
                print(e)
                exit(1)
            except Exception as e:
                try:
                    print(self.error(type(e), program, e))
                except KeyError:
                    print("An untracked exception has occurred\n Class: {}\n {}".format(type(e), e))
        else:
            return self.programs[program](**self.flags_dict)

    def flag_check(self, program, ap_args_object):
        """ Function checks if required flags were set for a given program

        :param ap_args_object: (ArgParse)   ArgParse object
        :param program: (str)	Name of program
        :return bool:
        """
        try:
            self.flags_dict = {flag: vars(ap_args_object)[flag] for flag in self.flags[program]}
        except KeyError as e:
            print(" ERROR: Invalid program name", str(e))
            exit(1)
        for flag in self.flags[program]:
            if self.flags_dict[flag] == "":
                return flag
        return None

    def run(self, args, debug=False):
        """ Primary function to call ProgramCaller and to run the script

        :param debug: (bool)    Set debug options to display full error Traceback
        :param args: (ArgParse.args)	ArgParse.args object
        """
        program = args.program
        failed = self.flag_check(program, args)
        if not failed:
            self.mapper(program, debug)
        else:
            self.error("INCORRECT_FLAGS", program, failed)


def db_validate(func):
    def wrapped(*args, **kwargs):
        func(*args, **kwargs)

    return func



# cython: language_level=3
import os
import re
from math import sqrt
# from io import StringIO
from random import choice
from sqlalchemy import text
from string import punctuation
from sqlalchemy.exc import OperationalError
from string import ascii_lowercase
from BioMetaDB.DBManagers.type_mapper import TypeMapper
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.DBOperations.update_existing_table import update_existing_table
from BioMetaDB.Exceptions.record_list_exceptions import ColumnNameNotFoundError


cdef extern from "Python.h":
    char* PyUnicode_AsUTF8(object unicode)


"""
Script holds functionality for handling data types associated with database records
Based on verbosity, will calculate standard deviation and averages of numerical values
And will provide info about most frequently occurring characters/words in character values

"""


cdef class RecordList(object):

    def __init__(self, object db_session=None, object table_class=None, object cfg=None, bint compute_metadata=False, str query=None,
                 list records_list = [], bint truncate = False):
        """ Class for handling (mostly) user interfacing to inner classes

        :param db_session:
        :param table_class:
        :param compute_metadata:
        :param truncate:
        """
        self.sess = db_session
        self.TableClass = table_class
        self.cfg = cfg
        self._summary = None
        self.num_records = 0
        self.truncate = truncate
        if records_list:
            self.results = records_list
        else:
            self.results = None
            self.num_records_in_db = len(self.sess.query(self.TableClass).all())
        self.has_text = False
        if query and not records_list:
            self.query(query)
        if compute_metadata and not records_list:
            self._summary, self.num_records, self.has_text = self._gather_metadata()

    def columns_summary(self):
        """ Wrapper for returning columns in class as simple dictionary Name: SQLType

        :return:
        """
        # cdef object summary_string = StringIO()
        cdef list sorted_keys
        cdef str key
        cdef int longest_key
        sorted_keys = sorted(ClassManager.get_class_as_dict(self.cfg).keys())
        longest_key = max([len(key) for key in sorted_keys])
        # Pretty formatting
        print("*" * (longest_key + 30))
        print("\t\t{:>{longest_key}}\t{:<12s}".format("Table Name:", self.cfg.table_name,
                                                                         longest_key=longest_key))
        print("\t\t{:>{longest_key}s}".format("Database", longest_key=longest_key))
        # Get all columns
        for key in sorted_keys:
            print("\t\t{:>{longest_key}s}".format(key, longest_key=longest_key))
        print("-" * (longest_key + 30))
        # return summary_string.getvalue()

    def table_name_summary(self):
        """ Wrapper for returning columns in class as simple dictionary Name: SQLType

        :return:
        """
        # cdef object summary_string = StringIO()
        # Pretty formatting
        print(self.cfg.table_name)
        # return summary_string.getvalue()

    def columns(self):
        """ Returns dict of columns

        :return:
        """
        if self.TableClass.__name__ == 'evaluation':
            reorder = ["domain", "phylum", "_class", "_order", "family", "genus", "species"]
            all_vals = list(ClassManager.get_class_as_dict(self.cfg).keys())
            for val in reorder:
                all_vals.remove(val)
            return reorder + all_vals
        return sorted(list(ClassManager.get_class_as_dict(self.cfg).keys()))

    def __str__(self):
        return self.summarize()

    def __repr__(self):
        return self.summarize()

    def summarize(self):
        """ Returns metadata for list of records queried in list

        :return:
        """
        # cdef object summary_string = StringIO()
        cdef list sorted_keys
        cdef str key, out_key, _out_key
        cdef int longest_key, num_none
        cdef object val
        if self._summary is None:
            self._summary, self.num_records, self.has_text = self._gather_metadata()
        sorted_keys = self.columns()
        longest_key = max([len(key) for key in sorted_keys])
        if longest_key < 18:
            longest_key = 18
        # Pretty formatting
        print("*" * (longest_key + 75))
        # Display multiple records
        if self.num_records > 1:
            print("\t{:>{longest_key}}\t{:<12s}\n\t{:>{longest_key}}\t{}".format(
                "Table Name:",
                self.cfg.table_name,
                "Number of Records:",
                "{:>{len_num}}/{:<{len_db}}".format(str(self.num_records), str(self.num_records_in_db),
                                               len_num=len(str(self.num_records)),
                                               len_db=len(str(self.num_records_in_db))),
                longest_key=longest_key))
        # Display single record
        elif self.num_records == 1:
            print(str(self.results[0]))
            return  # summary_string.getvalue()
        # No records found
        else:
            print("\t{:>{longest_key}}".format("No records found", longest_key=longest_key))
            # Do not create summary info
            return  # summary_string.getvalue()
        # Metadata display column headers
        print("\t{:>{longest_key}}\t{:<20s}\t{:<12s}".format(
            "Database",
            "Average",
            "Std Dev",
            longest_key=longest_key
        ))
        # Build summary string
        for key in sorted_keys:
            if key in self._summary.keys():
                if type(self._summary[key]) == list:
                    print("\t{:>{longest_key}}\t{:<20.3f}\t{:<12.3f}".format(
                        key,
                        self._summary[key][0],
                        self._summary[key][3],
                    longest_key=longest_key))
        print("-" * (longest_key + 75))
        if self.has_text:
            longest_key = max([len(key) for key in sorted_keys if key in self._summary.keys() and type(self._summary[key]) == dict])
            if longest_key < 18:
                longest_key = 18
            print("\n\t{:>{longest_key}}\t{:<20s}\t{:<10s}\t{:<12s}".format(
                "Database",
                "Most Frequent",
                "Number",
                "Total Count",
                longest_key=longest_key
            ))
            for key in sorted_keys:
                if key in self._summary.keys() and type(self._summary[key]) == dict:
                    num_none = self._summary[key].get("None", 0)
                    if num_none > 0:
                        del self._summary[key]["None"]
                    out_key = _out_key = max((self._summary[key].items() or {"n/a":0}.items()),
                                             key=lambda x : x[1])[0]
                    if out_key and len(out_key) > 16:
                        out_key = out_key[:17] + "..."
                    val = self._summary[key].get(_out_key, None)
                    print("\t{:>{longest_key}}\t{:<20s}\t{:<10d}\t{:<12.0f}".format(
                        str(key), (out_key if self.num_records == 1 or out_key != "n/a" else 'nil'),
                        (val if val and val != 1 else  1), self.num_records - num_none, longest_key=longest_key))
            print(("-" * (longest_key + 75)) + "\n")
        # return summary_string.getvalue()

    def _gather_metadata(self):
        """ Protected method to collect data, averages, and standard deviations

        :return:
        """
        cdef dict summary_data = {}, string_data = {}
        cdef int num_records, count
        cdef str column, val
        cdef list column_keys, vals
        cdef object record
        cdef bint has_text = False
        cdef object found_type, obj
        cdef dict data_types = {_k: TypeMapper.string_to_py_type[v]
                                for _k,v in ClassManager.get_class_as_dict(self.cfg).items()}
        if self.results is None:
             self.query()
        num_records = self.num_records
        column_keys = list(self.columns())
        if self.truncate:
            get_fxn = self._get_truncated
        else:
            get_fxn = getattr
        for record in self.results:
            # if record._id in self.ignore_ids:
            #     continue
            for column in column_keys:
                found_type = data_types[column]
                obj = getattr(record, column, None)
                if column not in summary_data.keys():
                    if obj is not None and obj != "None" and found_type in (int, float):
                        summary_data[column] = []
                        # Gather portion for average calculation
                        summary_data[column].append((float(obj) / float(num_records) if num_records > 0 else float(obj)))
                        # Gather portion for running sum
                        summary_data[column].append(float(obj))
                        # Gather portion for running sq sum
                        summary_data[column].append(float(obj) ** 2)
                        summary_data[column].append(0.0)
                    elif found_type in (str, bool):
                        has_text = True
                        # Gather count
                        val = str(getattr(record, column, ''))
                        vals = RecordList._correct_value((val if val != '' else "None"))
                        summary_data[column] = {}
                        for _v in vals:
                            if self.truncate:
                                _v = _v.split(" ")[0]
                            summary_data[column][_v] = 1
                else:
                    if obj is not None and obj != "None" and found_type in (int, float):
                        # Gather portion for average calculation
                        summary_data[column][0] += (float(obj) / float(num_records) if num_records > 0 else float(obj))
                        # Gather portion for running sum
                        summary_data[column][1] += float(obj)
                        # Gather portion for running sq sum
                        summary_data[column][2] += float(float(obj) ** 2)
                    elif found_type in (str, bool):
                        # Gather count
                        val = str(getattr(record, column, ""))
                        vals = RecordList._correct_value((val if val != '' else "None"))
                        for _v in vals:
                            if self.truncate:
                                _v = _v.split(" ")[0]
                            count = summary_data[column].get(_v, 0)
                            summary_data[column][_v] = count + 1
        # Determine standard deviation values
        if num_records > 1:
            for column in summary_data.keys():
                if type(summary_data[column]) != dict:
                # print(summary_data[column][1], summary_data[column][2])
                    try:
                        summary_data[column][3] = sqrt((summary_data[column][2] -
                                                        ((summary_data[column][1] ** 2) / num_records)) / (num_records - 1))
                    except ValueError:
                        summary_data[column][3] = -1
        return summary_data, num_records, has_text

    def query(self, *args):
        """ Wrapper function for querying database. Converts text argument to query statement

        :param args:
        :return:
        """
        cdef object record
        self._clear_prior_metadata()
        # Query passed
        if len(args) > 0:
            # Attempt query
            try:
                self.results = self._map_query(*args)
            # Column name not found
            except OperationalError:
                return
                # raise ColumnNameNotFoundError
        # Default get all
        else:
            self.results = self.sess.query(self.TableClass).all()
        self.num_records = len(self.results)
        return

    def _map_query(self, *args):
        """ Method calls 'convenience queries'

        :param args:
        :return:
        """
        cdef str query = "", col
        new_args = []
        for arg in args:
            new_args.append(arg)
            new_args[-1] = new_args[-1].rstrip(" ")
            new_args[-1] = new_args[-1].lstrip(" ")
        args = new_args
        # At least one annotation
        if "annotated" in args[0] and "unannotated" not in args[0]:
            for col in self.columns():
                query += "(%s IS NOT NULL AND %s != 'hypothetical protein') OR " % (col, col)
            args[0] = args[0].replace("annotated", query[:-4])
            return self.sess.query(self.TableClass).filter(text(*args)).all()
        # All unannotated
        if "unannotated" in args[0]:
            for col in self.columns():
                query += "(%s IS NULL OR %s == 'hypothetical protein') AND " % (col, col)
            args[0] = args[0].replace("unannotated", query[:-5])
            return self.sess.query(self.TableClass).filter(text(*args)).all()
        # High quality, non-redundant
        if 'hqnr' in args[0]:
            args[0] = args[0].replace("hqnr", "is_non_redundant AND is_complete AND NOT is_contaminated")
            return self.sess.query(self.TableClass).filter(text(*args)).all()
        # Individual annotations
        for val in args[0].split(" "):
            if "_annot" in val:
                _v = val.replace("_annot", "")
                args[0] = args[0].replace(val, "(%s IS NOT NULL AND %s != 'hypothetical protein')" % (_v, _v))
                # return self.sess.query(self.TableClass).filter(text(*args)).all()
            elif "_unannot" in val:
                _v = val.replace("_unannot", "")
                args[0] = args[0].replace(val, "(%s IS NULL OR %s == 'hypothetical protein')" % (_v, _v))
        return self.sess.query(self.TableClass).filter(text(*args)).all()


    # TODO Needs to be tested w/ tables that populate non-null set of db records
    def join(self, RecordList other):
        """ Method will create view of final

        :param other:
        :return:
        """
        try:
            self.results = self.sess.query(self.TableClass).join(other.TableClass, self.TableClass._id == other.TableClass._id).all()
        # Column name not found
        except OperationalError:
            raise ColumnNameNotFoundError
        return self

    def _clear_prior_metadata(self):
        """ Protected member will clear existing data
        To be called with query()

        :return:
        """
        self._summary, self.num_records = None, 0

    def find_column(self, str possible_column):
        """ Searches available columns for user-passed value
        Will attempt to look for values that may result from punctuation changes

        :return:
        """
        return RecordList._regex_search(possible_column, list(self.columns()))

    def __next__(self):
        """ Returns self as iterator

        :return:
        """
        return self

    def __contains__(self, item):
        return item._id in [record._id for record in self.results]

    def __iter__(self):
        """ Iterator yields next value in self.results

        :return:
        """
        cdef int i
        for i in range(self.num_records):
            yield self.results[i]

    def __getitem__(self, item):
        """ Allows class to be indexed and searched

        :param item:
        :return:
        """
        cdef object record
        if type(item) == int:
            assert item < self.num_records, "Index must be less than length"
            return self.results[item]
        elif type(item) == slice:
            if type(item.start) is None:
                item.start = 0
            if type(item.stop) is None:
                item.stop = -1
            if type(item.step) is None:
                item.start = 1
            return self.results[item.start : item.stop : item.step]
        elif type(item) == str:
            for record in self.results:
                if record._id == item:
                    return record
        return None

    def __len__(self):
        """ Accessible through len()

        :return:
        """
        return self.num_records

    def keys(self):
        """ Returns list of ids

        :return:
        """
        cdef object record
        if self.results:
            return [record._id for record in self.results]
        return None

    def values(self):
        """ Returns list of records

        :return:
        """
        cdef object record
        return [record for record in self.results]

    def items(self):
        """ Returns tuple of (key, value) pairs

        :return:
        """
        cdef object record
        return [(record._id, record) for record in self.results]

    def save(self):
        """ Calls the session's commit function to store changes

        :return:
        """
        self.sess.commit()
        return self

    def write_records(self, str output_dir):
        """ Method will write all records that are in the current db view to a user-provided
         output file

        :return:
        """
        cdef object R, W
        cdef bytes _line
        cdef object record
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        for record in self.results:
            W = open(os.path.join(output_dir, record._id), "wb")
            R = open(record.full_path(), "rb")
            try:
                for _line in R:
                    W.write(_line[:-1] + b" " + PyUnicode_AsUTF8(self._add_annotation(record)) + b"\n")
                    _line = next(R)
                    while _line[0] != b">":
                        W.write(_line)
                        _line = next(R)
            except StopIteration:
                continue
            R.close()
        W.close()

    def _add_annotation(self, object record):
        """ Protected method will search for annotation (via priority list) and add as fasta description

        :param record:
        :return:
        """
        cdef str annot, located_annot
        cdef list priority_list = RecordList._annotation_priority()
        if self.truncate:
            get_fxn = self._get_truncated
        else:
            get_fxn = getattr
        for annot in priority_list:
            located_annot = str(get_fxn(record, annot, ""))
            if located_annot is not None and located_annot not in ("", "None"):
                return located_annot
        if self.cfg is not None:
            for annot in set(self.columns()) - set(priority_list):
                located_annot = str(get_fxn(record, annot, ""))
                if located_annot is not None and located_annot not in ("", "None"):
                    return located_annot
        return ""

    def _get_truncated(self, object record, str annotation, str default=""):
        """ Protected method for returning truncated annotation, as needed

        :param annotation:
        :param default:
        :return:
        """
        # Prokka does not work well here
        # Do not truncate
        if annotation == "prokka":
            return str(getattr(record, annotation, default))
        cdef str trunc_annot = str(getattr(record, annotation, default)).split(" ")[0]
        if trunc_annot == "hypothetical":
            return "hypothetical protein"
        return trunc_annot

    def write_tsv(self, str output_file, str delim = "\t", list col_list = []):
        """ Outputs all record metadata in current db view to file

        :param output_file:
        :param delim:
        :param col_list:
        :return:
        """
        if self.truncate:
            get_fxn = self._get_truncated
        else:
            get_fxn = getattr
        if self.results is None:
             self.query()
        cdef list cols
        cdef object W
        cdef object record
        cdef str col
        # Write header
        if len(self.results) > 0:
            if col_list:
                cols = col_list
            else:
                cols = sorted(self.columns())
            W = open(output_file, "w")
            W.write("_id" + delim)
            W.write(delim.join(cols) + "\n")
            for record in self.results:
                W.write(record._id)
                for col in cols:
                    W.write(delim + str(get_fxn(record, col, "None")).strip())
                W.write("\n")

    def update(self, object data=None, str directory_name="None", bint silent=False, bint integrity_cancel=False):
        """ Wrapper function to update project. In experimental mode.

        :param data:
        :param directory_name:
        :param silent:
        :param integrity_cancel:
        :return:
        """
        # No valid info passed
        if data is None:
            return
        # Function call for .tsv file
        cdef str file_name = "%s.tmp" % "".join([choice(ascii_lowercase) for i in range(32)])
        data.to_file(file_name)
        update_existing_table(self.cfg.working_dir, self.TableClass.__name__, directory_name, file_name, "None", silent, integrity_cancel)
        data.delete_file()

    @staticmethod
    def _annotation_priority():
        return ["ko", "merops", "cazy", "merops_pfam", "prokka",
                "panther", "sfld", "tigrfam"]

    @staticmethod
    def _regex_search(str possible_column, list search_list):
        """ Protected method to search a list of values for a given string

        :param possible_column:
        :param search_list:
        :return:
        """
        cdef object r
        cdef set possible_columns
        cdef set cols_in_db = set(search_list)
        cdef dict db_cols = {col: col for col in cols_in_db}
        cdef set unusable_punctuation = set(punctuation) - set("_")
        cdef str punct
        cdef str tmp = possible_column
        r = re.compile(r"%s" % possible_column)
        # Exact matches
        possible_columns = set(filter(r.findall, cols_in_db))
        # Stored column name has punctuation, user value does not
        for punct in punctuation:
            db_cols = {col.replace(punct, ""): col for col in cols_in_db}
            possible_columns = possible_columns.union(set(db_cols[col] for col in filter(r.findall, db_cols.keys())))
            db_cols = {col.replace(punct, "_"): col for col in cols_in_db}
            possible_columns = possible_columns.union(set(db_cols[col] for col in filter(r.findall, db_cols.keys())))
        # User value has punctuation, stored column name does not
        for punct in unusable_punctuation:
            tmp = possible_column.replace(punct, "")
            r = re.compile(tmp)
            possible_columns = possible_columns.union(set(filter(r.findall, cols_in_db)))
            tmp = possible_column.replace(punct, "_")
            r = re.compile(tmp)
            possible_columns = possible_columns.union(set(filter(r.findall, cols_in_db)))
        # Return all possible values
        return list(possible_columns)

    @staticmethod
    def _correct_value(str value):
        """ Protected member function will split off annotation from %s-%s:::%s;;; paradigm

        :param value:
        :return:
        """
        cdef list return_list = [], vals
        cdef str val, _v
        if ";;;" in value:
            vals = [_v for _v in value.split(";;;") if _v != '']
            for val in vals:
                if ":::" in val:
                    _v = val.split(":::")[1]
                    if _v != '':
                        return_list.append(_v)
                else:
                    return_list.append(val)
            return return_list
        return [value,]
